import json
import logging
from typing import List, Dict, Any

#import httpx as requests
from httpx import AsyncClient as requests
from urllib.parse import urljoin
from datetime import datetime
import pytz

from . import xts_exception as ex

log = logging.getLogger(__name__)


class XTSCommon:
    """Base variables class"""
    def __init__(self, token=None, userID=None, isInvestorClient=None):

        self.token = token
        self.userID = userID
        self.isInvestorClient = isInvestorClient


class XTSConnect(XTSCommon):
    """The XTS Connect API wrapper class."""
    # Constants
    # Products
    PRODUCT_MIS = "MIS"
    PRODUCT_NRML = "NRML"
    PRODUCT_CNC = "CNC" # Added as per request from Jatin.

    # Order types
    ORDER_TYPE_MARKET = "MARKET"
    ORDER_TYPE_LIMIT = "LIMIT"
    ORDER_TYPE_STOPMARKET = "STOPMARKET"
    ORDER_TYPE_STOPLIMIT = "STOPLIMIT"

    # Transaction type
    TRANSACTION_TYPE_BUY = "BUY"
    TRANSACTION_TYPE_SELL = "SELL"

    # Squareoff mode
    SQUAREOFF_DAYWISE = "DayWise"
    SQUAREOFF_NETWISE = "Netwise"

    # Squareoff position quantity types
    SQUAREOFFQUANTITY_EXACTQUANTITY = "ExactQty"
    SQUAREOFFQUANTITY_PERCENTAGE = "Percentage"

    # Validity
    TimeinForce_GTC = "GTC"
    TimeinForce_IOC = "IOC"
    TimeinForce_FOK = "FOK"
    TimeinForce_GTD = "GTD"
    TimeinForce_DAY = "DAY"
    TimeinForce_AT_THE_OPEN = "AT_THE_OPEN"
    TimeinForce_AT_THE_CLOSE = "AT_THE_CLOSE"

    # Exchange Segments
    EXCHANGE_NSECM = "NSECM"
    EXCHANGE_NSEFO = "NSEFO"
    EXCHANGE_NSECD = "NSECD"
    EXCHANGE_MCXFO = "MCXFO"
    EXCHANGE_BSECM = "BSECM"
    EXCHANGE_BSEFO = "BSEFO"

    # URIs to various calls
    _routes = {
        # Interactive API endpoints
        "interactive.prefix": "interactive",
        "user.login": "/interactive/user/session",
        "user.logout": "/interactive/user/session",
        "user.profile": "/interactive/user/profile",
        "user.balance": "/interactive/user/balance",

        "orders": "/interactive/orders",
        "trades": "/interactive/orders/trades",
        "order.status": "/interactive/orders",
        "order.place": "/interactive/orders",
        "bracketorder.place": "/interactive/orders/bracket",
	    "bracketorder.modify": "/interactive/orders/bracket",
        "bracketorder.cancel": "/interactive/orders/bracket",
        "order.place.cover": "/interactive/orders/cover",
        "order.exit.cover": "/interactive/orders/cover",
        "order.modify": "/interactive/orders",
        "order.cancel": "/interactive/orders",
        "order.cancelall": "/interactive/orders/cancelall",
        "order.history": "/interactive/orders",

        "portfolio.positions": "/interactive/portfolio/positions",
        "portfolio.holdings": "/interactive/portfolio/holdings",
        "portfolio.positions.convert": "/interactive/portfolio/positions/convert",
        "portfolio.squareoff": "/interactive/portfolio/squareoff",
	    "portfolio.dealerpositions": "interactive/portfolio/dealerpositions",
	    "order.dealer.status": "/interactive/orders/dealerorderbook",
	    "dealer.trades": "/interactive/orders/dealertradebook",

        # Market API endpoints
        "marketdata.prefix": "apimarketdata",
        "market.login": "/apimarketdata/auth/login",
        "market.logout": "/apimarketdata/auth/logout",

        "market.config": "/apimarketdata/config/clientConfig",

        "market.instruments.master": "/apimarketdata/instruments/master",
        "market.instruments.subscription": "/apimarketdata/instruments/subscription",
        "market.instruments.unsubscription": "/apimarketdata/instruments/subscription",
        "market.instruments.ohlc": "/apimarketdata/instruments/ohlc",
        "market.instruments.indexlist": "/apimarketdata/instruments/indexlist",
        "market.instruments.quotes": "/apimarketdata/instruments/quotes",

        "market.search.instrumentsbyid": '/apimarketdata/search/instrumentsbyid',
        "market.search.instrumentsbystring": '/apimarketdata/search/instruments',

        "market.instruments.instrument.series": "/apimarketdata/instruments/instrument/series",
        "market.instruments.instrument.equitysymbol": "/apimarketdata/instruments/instrument/symbol",
        "market.instruments.instrument.futuresymbol": "/apimarketdata/instruments/instrument/futureSymbol",
        "market.instruments.instrument.optionsymbol": "/apimarketdata/instruments/instrument/optionsymbol",
        "market.instruments.instrument.optiontype": "/apimarketdata/instruments/instrument/optionType",
        "market.instruments.instrument.expirydate": "/apimarketdata/instruments/instrument/expiryDate"
    }

    def __init__(self,
                 apiKey,
                 secretKey,
                 source,
                 root,
                 debug=False,
                 timeout=1200, # chnaged from 7 to 1200, around 20 minutes.
                 pool=None,
                 disable_ssl=True):
        """
        Initialise a new XTS Connect client instance.

        - `apikey` is the key issued to you
        - `root` is the API end point root. Unless you explicitly
        want to send API requests to a non-default endpoint, this
        can be ignored.
        - `debug`, if set to True, will serialise and print requests
        and responses to stdout.
        - `timeout` is the time (seconds) for which the API client will wait for
        a request to complete before it fails. Defaults to 7 seconds
        - `pool` is manages request pools. It takes a dict of params accepted by HTTPAdapter
        - `disable_ssl` disables the SSL verification while making a request.
        If set requests won't throw SSLError if its set to custom `root` url without SSL.
        """
        self.debug = debug
        self.apiKey = apiKey
        self.secretKey = secretKey
        self.source = source
        self.disable_ssl = disable_ssl
        self.root = root 
        self.timeout = timeout
        self.last_login_time = None   

        super().__init__()

        # Create requests session only if pool exists. Reuse session
        # for every request. Otherwise create session for each request
        if pool:
            self.reqsession = requests.Session()
            reqadapter = requests.adapters.HTTPAdapter(**pool)
            self.reqsession.mount("https://", reqadapter)
        else:
            self.reqsession = requests()

        # disable requests SSL warning
        #requests.packages.urllib3.disable_warnings()

    def _set_common_variables(self, access_token,userID, isInvestorClient):
        """
        Set the `access_token` received after a successful authentication.
        HELPER FUNCTION, DO NOT CALL DIRECTLY.
        """
        super().__init__(access_token,userID, isInvestorClient)

    def _login_url(self):
        """Get the remote login url to which a user should be redirected to initiate the login flow."""
        return self.root +  "/user/session"

    async def interactive_login(self):
        """
        Send the login url to which a user should receive the token.
        ```
        import os
        API_key = os.getenv("API_KEY")
        API_secret = os.getenv("API_SECRET")
        API_source = os.getenv("API_SOURCE")
        API_root = os.getenv("API_URL")
        """"""""""""""""""""""""""""""""""""""""""
            |DataFrame for Cash Market|
        """"""""""""""""""""""""""""""""""""""""""
        from xts_api_client.xts_connect_async import XTSConnect
        import asyncio

        async def main():
            xt_market_data = XTSConnect(
            apiKey = API_key,
            secretKey = API_secret,
            source = API_source,
            root = API_root
            )
            responce_login = await xt_market_data.marketdata_login()
            print(f"Loggin In: {responce_login}")
            
            await xt_market_data.marketdata_logout()
        if __name__ == "__main__":
            asyncio.run(main())
        ```
        """
        try:
            params = {
                "appKey": self.apiKey,
                "secretKey": self.secretKey,
                "source": self.source
            }
            response = await self._post("user.login", params)

            if "token" in response['result']:
                self._set_common_variables(response['result']['token'], response['result']['userID'],
                                           response['result']['isInvestorClient'])
                self._last_login_time = datetime.now(pytz.timezone("Asia/Kolkata"))
            return response
        except Exception as e:
            return response['description']

    async def get_order_book(self, clientID=None):
        """
        Request Order book gives states of all the orders placed by an user.
        IMPORTANT: THIS WILL ONLY WORK AFTER LOGGING IN USING `interactive_login` METHOD.
        ```
        import asyncio
        from xts_api_client.xts_connect_async import XTSConnect
        import os

        API_key = os.getenv("INTERACTIVE_API_KEY")
        API_secret = os.getenv("INTERACTIVE_API_SECRET")
        API_source = os.getenv("API_SOURCE")
        API_root = os.getenv("API_URL")


        xt_interactive = XTSConnect(
        apiKey = API_key,
        secretKey = API_secret,
        source = API_source,
        root = API_root
        )


        async def main():
            await xt_interactive.interactive_login()
            
            respnse_get_order_book_list = await xt_interactive.get_order_book()
            print(f"Order Book: {respnse_get_order_book_list['result']}")
            
            await xt_interactive.interactive_logout()
        if __name__ == "__main__":
            asyncio.run(main())

        ```
        """
        try:
            params = {}
            if not self.isInvestorClient:
                params['clientID'] = "*****"
            else:
                params['clientID'] = clientID
            response = await self._get("order.status", params)
            return response
        except Exception as e:
            return response['description']
		
    async def get_dealer_orderbook(self, clientID=None):
        """
        Request Order book gives states of all the orders placed by an user.
        IMPORTANT: THIS WILL ONLY WORK AFTER LOGGING IN USING `interactive_login` METHOD.
        ```
        import asyncio
        from xts_api_client.xts_connect_async import XTSConnect
        import os

        API_key = os.getenv("INTERACTIVE_API_KEY")
        API_secret = os.getenv("INTERACTIVE_API_SECRET")
        API_source = os.getenv("API_SOURCE")
        API_root = os.getenv("API_URL")


        xt_interactive = XTSConnect(
        apiKey = API_key,
        secretKey = API_secret,
        source = API_source,
        root = API_root
        )

        async def main():
            await xt_interactive.interactive_login()
            
            respnse_get_dealer_orderbook = await xt_interactive.get_dealer_orderbook()
            list_of_order_in_dealer_orderbook = respnse_get_dealer_orderbook['result']
            print(f"Dealer Order Book: {list_of_order_in_dealer_orderbook}")
            
            await xt_interactive.interactive_logout()
        if __name__ == "__main__":
            asyncio.run(main())

        ```
        """
        try:
            params = {}
            if not self.isInvestorClient:
                params['clientID'] = "*****"
            else:
                params['clientID'] = clientID
            response = await self._get("order.dealer.status", params)
            return response
        except Exception as e:
            return response['description']

    async def place_order(self,
                    exchangeSegment,
                    exchangeInstrumentID,
                    productType,
                    orderType,
                    orderSide,
                    timeInForce,
                    disclosedQuantity,
                    orderQuantity,
                    limitPrice,
                    stopPrice,
                    orderUniqueIdentifier,
                    clientID=None
                    ):
        """
        
        """
        try:

            params = {
                "exchangeSegment": exchangeSegment,
                "exchangeInstrumentID": exchangeInstrumentID,
                "productType": productType,
                "orderType": orderType,
                "orderSide": orderSide,
                "timeInForce": timeInForce,
                "disclosedQuantity": disclosedQuantity,
                "orderQuantity": orderQuantity,
                "limitPrice": limitPrice,
                "stopPrice": stopPrice,
                "orderUniqueIdentifier": orderUniqueIdentifier
            }

            if not self.isInvestorClient:
                params['clientID'] = '*****' # Jatin
            else:
                params['clientID'] = self.userID

            response = await self._post('order.place', json.dumps(params))
            return response
        except Exception as e:
            return response['description']
        
    async def place_bracketorder(self,
                    exchangeSegment,
                    exchangeInstrumentID,
                    orderType,
                    orderSide,
                    disclosedQuantity,
                    orderQuantity,
                    limitPrice,
                    squarOff,
                    stopLossPrice,
	                trailingStoploss,
                    isProOrder,
                    orderUniqueIdentifier,
                     ):
        """To place a bracketorder"""
        try:

            params = {
                "exchangeSegment": exchangeSegment,
                "exchangeInstrumentID": exchangeInstrumentID,
                "orderType": orderType,
                "orderSide": orderSide,
                "disclosedQuantity": disclosedQuantity,
                "orderQuantity": orderQuantity,
                "limitPrice": limitPrice,
                "squarOff": squarOff,
                "stopLossPrice": stopLossPrice,
                "trailingStoploss": trailingStoploss,
                "isProOrder": isProOrder,
             "orderUniqueIdentifier": orderUniqueIdentifier
            }
            response = await self._post('bracketorder.place', json.dumps(params))
            print(response)
            return response
        except Exception as e:
            return response['description']

    async def get_profile(self, clientID=None):
        """
        THIS IS NOT WORKING AS OF NOW.
        """
        try:
            params = {}
            if not self.isInvestorClient:
                params['clientID'] = "*****"
            else:
                params['clientID'] = clientID

            response = await self._get('user.profile', params)
            return response
        except Exception as e:
            return response['description']

    async def get_balance(self, clientID=None):
        """
        Using session token user can access his balance stored with the broker.
        IMPORTANT: THIS WILL ONLY WORK AFTER LOGGING IN USING `interactive_login` METHOD.
        ```
        import asyncio
        from xts_api_client.xts_connect_async import XTSConnect
        import os

        API_key = os.getenv("INTERACTIVE_API_KEY")
        API_secret = os.getenv("INTERACTIVE_API_SECRET")
        API_source = os.getenv("API_SOURCE")
        API_root = os.getenv("API_URL")


        xt_interactive = XTSConnect(
        apiKey = API_key,
        secretKey = API_secret,
        source = API_source,
        root = API_root
        )

        async def main():
            await xt_interactive.interactive_login()
            
            respnse_get_profile = await xt_interactive.get_balance()
            print(f"Cash Available: {respnse_get_profile['result']['BalanceList'][0]['limitObject']['RMSSubLimits']['cashAvailable']}")
            print(f"Net Margin Available: {respnse_get_profile['result']['BalanceList'][0]['limitObject']['RMSSubLimits']['netMarginAvailable']}")
            
            await xt_interactive.interactive_logout()
        if __name__ == "__main__":
            asyncio.run(main())
        ```
        """
        try:
            params = {}
            if not self.isInvestorClient:
                params['clientID'] = "*****"
            else:
                params['clientID'] = clientID
            response = await self._get('user.balance', params)
            return response
        except Exception as e:
            return response['description']
   
    async def modify_order(self,
                     appOrderID,
                     modifiedProductType,
                     modifiedOrderType,
                     modifiedOrderQuantity,
                     modifiedDisclosedQuantity,
                     modifiedLimitPrice,
                     modifiedStopPrice,
                     modifiedTimeInForce,
                     orderUniqueIdentifier,
                     clientID=None
                     ):
        """The facility to modify your open orders by allowing you to change limit order to market or vice versa,
        change Price or Quantity of the limit open order, change disclosed quantity or stop-loss of any
        open stop loss order. """
        try:
            appOrderID = int(appOrderID)
            params = {
                'appOrderID': appOrderID,
                'modifiedProductType': modifiedProductType,
                'modifiedOrderType': modifiedOrderType,
                'modifiedOrderQuantity': modifiedOrderQuantity,
                'modifiedDisclosedQuantity': modifiedDisclosedQuantity,
                'modifiedLimitPrice': modifiedLimitPrice,
                'modifiedStopPrice': modifiedStopPrice,
                'modifiedTimeInForce': modifiedTimeInForce,
                'orderUniqueIdentifier': orderUniqueIdentifier
            }

            if not self.isInvestorClient:
                params['clientID'] = '*****' # Jatin, again :-)
            else:
                params['clientID'] = self.userID

            response = await self._put('order.modify', json.dumps(params))
            return response
        except Exception as e:
            return response['description']

    async def get_trade(self, clientID=None):
        """
        Trade book returns a list of all trades executed on a particular day , that were placed by the user . The
        trade book will display all filled and partially filled orders.
        IMPORTANT: THIS WILL ONLY WORK AFTER LOGGING IN USING `interactive_login` METHOD.
        ```
        import asyncio
        from xts_api_client.xts_connect_async import XTSConnect
        import os

        API_key = os.getenv("INTERACTIVE_API_KEY")
        API_secret = os.getenv("INTERACTIVE_API_SECRET")
        API_source = os.getenv("API_SOURCE")
        API_root = os.getenv("API_URL")


        xt_interactive = XTSConnect(
        apiKey = API_key,
        secretKey = API_secret,
        source = API_source,
        root = API_root
        )

        async def main():
            await xt_interactive.interactive_login()
            
            respnse_get_trade = await xt_interactive.get_trade()
            list_of_trade = respnse_get_trade['result']
            print(f"Trade List: {list_of_trade}")
            
            await xt_interactive.interactive_logout()
        if __name__ == "__main__":
            asyncio.run(main())

        ```
        """
        try:
            params = {}
            if not self.isInvestorClient:
                params['clientID'] = "*****"
            else:
                params['clientID'] = clientID
            response = await self._get('trades', params)
            return response
        except Exception as e:
            return response['description']

    async def get_dealer_tradebook(self, clientID=None):
        """
        Trade book returns a list of all trades executed on a particular day , that were placed by the user . The
        trade book will display all filled and partially filled orders.
        IMPORTANT: THIS WILL ONLY WORK AFTER LOGGING IN USING `interactive_login` METHOD.
        ```
        import asyncio
        from xts_api_client.xts_connect_async import XTSConnect
        import os

        API_key = os.getenv("INTERACTIVE_API_KEY")
        API_secret = os.getenv("INTERACTIVE_API_SECRET")
        API_source = os.getenv("API_SOURCE")
        API_root = os.getenv("API_URL")


        xt_interactive = XTSConnect(
        apiKey = API_key,
        secretKey = API_secret,
        source = API_source,
        root = API_root
        )

        async def main():
            await xt_interactive.interactive_login()
            
            respnse_get_dealer_tradebook = await xt_interactive.get_dealer_tradebook()
            list_of_trade = respnse_get_dealer_tradebook['result']
            print(f"Trade List: {list_of_trade}")
            
            await xt_interactive.interactive_logout()
        if __name__ == "__main__":
            asyncio.run(main())

        ```
        """
        try:
            params = {}
            if not self.isInvestorClient:
                params['clientID'] = "*****"
            else:
                params['clientID'] = clientID
            response = await self._get('dealer.trades', params)
            return response
        except Exception as e:
            return response['description']
		
    async def get_holding(self, clientID=None):
        """
        Holdings API call enable users to check their long term holdings with the broker.
        IMPORTANT: THIS WILL ONLY WORK AFTER LOGGING IN USING `interactive_login` METHOD.
        ```
        import asyncio
        from xts_api_client.xts_connect_async import XTSConnect
        import os

        API_key = os.getenv("INTERACTIVE_API_KEY")
        API_secret = os.getenv("INTERACTIVE_API_SECRET")
        API_source = os.getenv("API_SOURCE")
        API_root = os.getenv("API_URL")


        xt_interactive = XTSConnect(
        apiKey = API_key,
        secretKey = API_secret,
        source = API_source,
        root = API_root
        )

        async def main():
            await xt_interactive.interactive_login()
            
            respnse_get_holding = await xt_interactive.get_holding()
            holding = respnse_get_holding['result']
            print(f"Trade List: {holding}")
            
            await xt_interactive.interactive_logout()
        if __name__ == "__main__":
            asyncio.run(main())

        ```
        """
        try:
            params = {}
            if not self.isInvestorClient:
                params['clientID'] = "*****"
            else:
                params['clientID'] = clientID

            response = await self._get('portfolio.holdings', params)
            return response
        except Exception as e:
            return response['description']

    async def bracketorder_cancel(self, appOrderID, clientID=None):
        """This API can be called to cancel any open order of the user by providing correct appOrderID matching with
        the chosen open order to cancel. """
        try:
            params = {'boEntryOrderId': int(appOrderID)}
            if not self.isInvestorClient:
                params['clientID'] = "*****"
            else:
                params['clientID'] = clientID
            response = await self._delete('bracketorder.cancel', params)
            return response
        except Exception as e:
            return response['description']   
		
    async def get_dealerposition_netwise(self, clientID=None):
        """The positions API positions by net. Net is the actual, current net position portfolio."""
        try:
            params = {'dayOrNet': 'NetWise'}
            if not self.isInvestorClient:
                params['clientID'] = "*****"
            else:
                params['clientID'] = clientID
            response = await self._get('portfolio.dealerpositions', params)
            return response
        except Exception as e:
            return response['description']


           
    async def get_dealerposition_daywise(self, clientID=None):
        """The positions API returns positions by day, which is a snapshot of the buying and selling activity for
        that particular day."""
        try:
            params = {'dayOrNet': 'DayWise'}
            if not self.isInvestorClient:
                params['clientID'] = "*****"
            else:
                params['clientID'] = clientID
            response = await self._get('portfolio.dealerpositions', params)
            return response
        except Exception as e:
            return response['description']
		
    async def get_position_daywise(self, clientID=None):
	    
        """The positions API returns positions by day, which is a snapshot of the buying and selling activity for
        that particular day."""
        try:
            params = {'dayOrNet': 'DayWise'}
            if not self.isInvestorClient:
                params['clientID'] = "*****"
            else:
                params['clientID'] = clientID
            response = await self._get('portfolio.positions', params)
            return response
        except Exception as e:
            return response['description']

    async def get_position_netwise(self, clientID=None):
        """The positions API positions by net. Net is the actual, current net position portfolio."""
        try:
            params = {'dayOrNet': 'NetWise'}
            if not self.isInvestorClient:
                params['clientID'] = "*****"
            else:
                params['clientID'] = clientID
            response = await self._get('portfolio.positions', params)
            return response
        except Exception as e:
            return response['description']

    async def convert_position(self, exchangeSegment, exchangeInstrumentID, targetQty, isDayWise, oldProductType,
                         newProductType, clientID=None):
        """Convert position API, enable users to convert their open positions from NRML intra-day to Short term MIS or
        vice versa, provided that there is sufficient margin or funds in the account to effect such conversion """
        try:
            params = {
                'exchangeSegment': exchangeSegment,
                'exchangeInstrumentID': exchangeInstrumentID,
                'targetQty': targetQty,
                'isDayWise': isDayWise,
                'oldProductType': oldProductType,
                'newProductType': newProductType
            }
            if not self.isInvestorClient:
                params['clientID'] = "*****"
            else:
                params['clientID'] = clientID
            response = await self._put('portfolio.positions.convert', json.dumps(params))
            return response
        except Exception as e:
            return response['description']

    async def cancel_order(self, appOrderID, orderUniqueIdentifier, clientID=None):
        """This API can be called to cancel any open order of the user by providing correct appOrderID matching with
        the chosen open order to cancel. """
        try:
            params = {'appOrderID': int(appOrderID), 'orderUniqueIdentifier': orderUniqueIdentifier}
            if not self.isInvestorClient:
                params['clientID'] = '*****' # Jatin, returns
            else:
                params['clientID'] = self.userID
            response = await self._delete('order.cancel', params)
            return response
        except Exception as e:
            return response['description']
        
    async def cancelall_order(self, exchangeSegment, exchangeInstrumentID):
        """This API can be called to cancel all open order of the user by providing exchange segment and exchange instrument ID """
        try:
            params = {"exchangeSegment": exchangeSegment, "exchangeInstrumentID": exchangeInstrumentID}
            if not self.isInvestorClient:
                params['clientID'] = '*****' # Jatin, returns
            else:
                params['clientID'] = self.userID
            response = await self._post('order.cancelall', json.dumps(params))
            return response
        except Exception as e:
            return response['description']    

    async def place_cover_order(self, exchangeSegment, exchangeInstrumentID, orderSide,orderType, orderQuantity, disclosedQuantity,
                          limitPrice, stopPrice, orderUniqueIdentifier, clientID=None):
        """A Cover Order is an advance intraday order that is accompanied by a compulsory Stop Loss Order. This helps
        users to minimize their losses by safeguarding themselves from unexpected market movements. A Cover Order
        offers high leverage and is available in Equity Cash, Equity F&O, Commodity F&O and Currency F&O segments. It
        has 2 orders embedded in itself, they are Limit/Market Order Stop Loss Order """
        try:

            params = {'exchangeSegment': exchangeSegment, 'exchangeInstrumentID': exchangeInstrumentID,
                      'orderSide': orderSide, "orderType": orderType,'orderQuantity': orderQuantity, 'disclosedQuantity': disclosedQuantity,
                      'limitPrice': limitPrice, 'stopPrice': stopPrice, 'orderUniqueIdentifier': orderUniqueIdentifier}
            if not self.isInvestorClient:
                params['clientID'] = "*****"
            else:
                params['clientID'] = clientID
            response = await self._post('order.place.cover', json.dumps(params))
            return response
        except Exception as e:
            return response['description']

    async def exit_cover_order(self, appOrderID, clientID=None):
        """Exit Cover API is a functionality to enable user to easily exit an open stoploss order by converting it
        into Exit order. """
        try:

            params = {'appOrderID': appOrderID}
            if not self.isInvestorClient:
                params['clientID'] = "*****"
            else:
                params['clientID'] = clientID
            response = await self._put('order.exit.cover', json.dumps(params))
            return response
        except Exception as e:
            return response['description']

    async def squareoff_position(self, exchangeSegment, exchangeInstrumentID, productType, squareoffMode,
                           positionSquareOffQuantityType, squareOffQtyValue, blockOrderSending, cancelOrders,
                           clientID=None):
        """User can request square off to close all his positions in Equities, Futures and Option. Users are advised
        to use this request with caution if one has short term holdings. """
        try:

            params = {'exchangeSegment': exchangeSegment, 'exchangeInstrumentID': exchangeInstrumentID,
                      'productType': productType, 'squareoffMode': squareoffMode,
                      'positionSquareOffQuantityType': positionSquareOffQuantityType,
                      'squareOffQtyValue': squareOffQtyValue, 'blockOrderSending': blockOrderSending,
                      'cancelOrders': cancelOrders
                      }
            if not self.isInvestorClient:
                params['clientID'] = "*****"
            else:
                params['clientID'] = clientID
            response = await self._put('portfolio.squareoff', json.dumps(params))
            return response
        except Exception as e:
            return response['description']

    async def get_order_history(self, appOrderID, clientID=None):
        """Order history will provide particular order trail chain. This indicate the particular order & its state
        changes. i.e.Pending New to New, New to PartiallyFilled, PartiallyFilled, PartiallyFilled & PartiallyFilled
        to Filled etc """
        try:
            params = {'appOrderID': appOrderID}
            if not self.isInvestorClient:
                params['clientID'] = "*****"
            else:
                params['clientID'] = clientID
            response = await self._get('order.history', params)
            return response
        except Exception as e:
            return response['description']

    async def interactive_logout(self, clientID=None):
        """This call invalidates the session token and destroys the API session. After this, the user should go
        through login flow again and extract session token from login response before further activities. """
        try:
            params = {}
            if not self.isInvestorClient:
                params['clientID'] = "*****"
            else:
                params['clientID'] = clientID
            response = await self._delete('user.logout', params)
            self.token = None # Added this to reset token, so that we can login again.
            return response
        except Exception as e:
            return response['description']

    ########################################################################################################
    # Market data API
    ########################################################################################################

    async def marketdata_login(self):
        """Send the login url to which a user should receive the token."""
        try:
            params = {
                "appKey": self.apiKey,
                "secretKey": self.secretKey,
                "source": self.source
            }
            response = await self._post("market.login", params)

            if "token" in response['result']:
                self._set_common_variables(response['result']['token'], response['result']['userID'],False)
                self._last_login_time = datetime.now(pytz.timezone("Asia/Kolkata"))
            return response
        except Exception as e:
            return response['description']

    async def get_config(self):
        """Get the configuration of the client."""
        try:
            params = {}
            response = await self._get('market.config', params)
            return response
        except Exception as e:
            return response['description']

    async def get_quote(self, Instruments, xtsMessageCode, publishFormat):
        """Get the quote of the instrument."""
        try:

            params = {'instruments': Instruments, 'xtsMessageCode': xtsMessageCode, 'publishFormat': publishFormat}
            response = await self._post('market.instruments.quotes', json.dumps(params))
            return response
        except Exception as e:
            return response['description']

    async def send_subscription(
        self,
        Instruments: List[Dict[str, int]],
        xtsMessageCode: int
        ):
        """
        Sends a subscription request for the specified instruments using the given XTS message code.

        Parameters:
            Instruments (List[Dict[str, int]]): A list of instruments to subscribe to. 
                Each instrument should be a dictionary with:
                    - 'exchangeSegment' (int): Numeric code for the exchange segment.
                    - 'exchangeInstrumentID' (int): Unique instrument ID in that segment.

                Example:
                    instruments = [
                        {'exchangeSegment': 1, 'exchangeInstrumentID': 2885},
                        {'exchangeSegment': 1, 'exchangeInstrumentID': 22}
                    ]

            xtsMessageCode (int): XTS message code specifying the type of subscription.
                For example:
                -TouchlineEvent               = 1501,
                -MarketDepthEvent             = 1502,
                -IndexDataEvent               = 1504,
                -CandleDataEvent              = 1505,
                -InstrumentPropertyChangeEvent = 1105,
                -OpenInterestEvent            = 1510,
                -LTPEvent                     = 1512
            """
        try:
            params = {'instruments': Instruments, 'xtsMessageCode': xtsMessageCode}
            response = await self._post('market.instruments.subscription', json.dumps(params))
            return response
        except Exception as e:
            return response['description']

    async def send_unsubscription(
        self,
        Instruments: List[Dict[str, int]],
        xtsMessageCode: int
        ):
        """
        Sends an unsubscription request to stop receiving data for the specified instruments.

        Parameters:
            Instruments (List[Dict[str, int]]): A list of instruments to unsubscribe from. 
                Each instrument should be a dictionary containing:
                    - 'exchangeSegment' (int): Numeric code for the exchange segment.
                    - 'exchangeInstrumentID' (int): Unique instrument ID in that segment.

                Example:
                    instruments = [
                        {'exchangeSegment': 1, 'exchangeInstrumentID': 2885},
                        {'exchangeSegment': 1, 'exchangeInstrumentID': 22}
                    ]

            xtsMessageCode (int): XTS message code specifying the type of subscription to cancel.
                For example:
                -TouchlineEvent               = 1501,
                -MarketDepthEvent             = 1502,
                -IndexDataEvent               = 1504,
                -CandleDataEvent              = 1505,
                -InstrumentPropertyChangeEvent = 1105,
                -OpenInterestEvent            = 1510,
                -LTPEvent                     = 1512
        """
        try:
            params = {'instruments': Instruments, 'xtsMessageCode': xtsMessageCode}
            response = await self._put('market.instruments.unsubscription', json.dumps(params))
            return response
        except Exception as e:
            return response['description']

    async def get_master(
        self,
        exchangeSegmentList: List[int]
        ):
        """Get the master string."""
        try:
            params = {"exchangeSegmentList": exchangeSegmentList}
            response = await self._post('market.instruments.master', json.dumps(params))
            return response
        except Exception as e:
            return response['description']

    async def get_ohlc(
        self,
        exchangeSegment: int,
        exchangeInstrumentID: int,
        startTime: str,
        endTime: str,
        compressionValue: int):
        """
        Retrieves historical OHLC (Open, High, Low, Close) candle data for a given instrument.

        Parameters:
        exchangeSegment (int): Numeric identifier for the exchange segment. Supported values:
                - 1  = NSECM
                - 2  = NSEFO
                - 3  = NSECD
                - 4  = NSECO
                - 5  = SLBM
                - 7  = NIFSC
                - 11 = BSECM
                - 12 = BSEFO
                - 13 = BSECD
                - 14 = BSECO
                - 21 = NCDEX
                - 41 = MSECM
                - 42 = MSEFO
                - 43 = MSECD
                - 51 = MCXFO

        exchangeInstrumentID (int): Unique instrument ID for the given exchange segment.
            - For NSECM, use NSE instrument ID (e.g., 22 for NIFTY).
            - For BSECM, use BSE instrument ID (e.g., "526530").

        startTime (str): Start time for the OHLC data in the format "MMM DD YYYY HHMMSS", e.g., "Dec 02 2024 091500".
            - Time is in IST (Indian Standard Time).

        endTime (str): End time for the OHLC data in the format "MMM DD YYYY HHMMSS", e.g., "Dec 02 2024 133000".
            - Time is in IST and uses 24-hour format.

        compressionValue (int): Timeframe for each candle in minutes.
            - For example, 1 for 1-minute candles, 60 for hourly candles, etc.
        """
        try:
            params = {
                'exchangeSegment': exchangeSegment,
                'exchangeInstrumentID': exchangeInstrumentID,
                'startTime': startTime,
                'endTime': endTime,
                'compressionValue': compressionValue}
            response = await self._get('market.instruments.ohlc', params)
            return response
        except Exception as e:
            return response['description']

    async def get_series(
        self,
        exchangeSegment: int
        ):
        """ 
        Retrieves the series of instruments available on a specific exchange segment.
        Parameters:
        exchangeSegment (int): Numeric identifier for the exchange segment. Supported values include:
                - 1  = NSECM
                - 2  = NSEFO
                - 3  = NSECD
                - 4  = NSECO
                - 5  = SLBM
                - 7  = NIFSC
                - 11 = BSECM
                - 12 = BSEFO
                - 13 = BSECD
                - 14 = BSECO
                - 21 = NCDEX
                - 41 = MSECM
                - 42 = MSEFO
                - 43 = MSECD
                - 51 = MCXFO
        """
        try:
            params = {'exchangeSegment': exchangeSegment}
            response = await self._get('market.instruments.instrument.series', params)
            return response
        except Exception as e:
            return response['description']

    async def get_equity_symbol(
        self,
        exchangeSegment: int,
        series: str,
        symbol: str
        ):
        """ 
        Retrieves the full equity symbol for a given instrument based on the exchange segment, series, and trading symbol.

        Parameters:
        exchangeSegment (int): Numeric identifier for the exchange segment. Supported values include:
                - 1  = NSECM
                - 2  = NSEFO
                - 3  = NSECD
                - 4  = NSECO
                - 5  = SLBM
                - 7  = NIFSC
                - 11 = BSECM
                - 12 = BSEFO
                - 13 = BSECD
                - 14 = BSECO
                - 21 = NCDEX
                - 41 = MSECM
                - 42 = MSEFO
                - 43 = MSECD
                - 51 = MCXFO

        series (str): Series type for the equity, such as:
            - "EQ"  = Equity
            - "BE"  = Trade-to-trade segment
            - "BL", "BZ", etc., as applicable

        symbol (str): Trading symbol of the security, e.g., "RELIANCE", "TATAMOTORS", "INFY".
        
        """
        try:

            params = {'exchangeSegment': exchangeSegment, 'series': series, 'symbol': symbol}
            response = await self._get('market.instruments.instrument.equitysymbol', params)
            return response
        except Exception as e:
            return response['description']

    async def get_expiry_date(
        self,
        exchangeSegment: int,
        series: str,
        symbol: str
        ):
        """
        Retrieves the available expiry dates for a given instrument on the specified exchange segment.

        Parameters:
            exchangeSegment (int): Numeric identifier for the exchange segment. Accepted values include:
                - 1  = NSECM
                - 2  = NSEFO
                - 3  = NSECD
                - 4  = NSECO
                - 5  = SLBM
                - 7  = NIFSC
                - 11 = BSECM
                - 12 = BSEFO
                - 13 = BSECD
                - 14 = BSECO
                - 21 = NCDEX
                - 41 = MSECM
                - 42 = MSEFO
                - 43 = MSECD
                - 51 = MCXFO

            series (str): Series type, such as "FUT", "OPTIDX", "OPTSTK", etc.

            symbol (str): Trading symbol of the instrument, e.g., "NIFTY", "BANKNIFTY", "RELIANCE".
        """
        try:
            params = {'exchangeSegment': exchangeSegment, 'series': series, 'symbol': symbol}
            response = await self._get('market.instruments.instrument.expirydate', params)
            return response
        except Exception as e:
            return response['description']

    async def get_future_symbol(
        self,
        exchangeSegment: int,
        series: str,
        symbol: str,
        expiryDate: str
        ):
        """
        Retrieves the future symbol for the specified instrument parameters.

        Parameters:
            exchangeSegment (int): Numeric identifier for the exchange segment. Accepted values include:
                - 1  = NSECM
                - 2  = NSEFO
                - 3  = NSECD
                - 4  = NSECO
                - 5  = SLBM
                - 7  = NIFSC
                - 11 = BSECM
                - 12 = BSEFO
                - 13 = BSECD
                - 14 = BSECO
                - 21 = NCDEX
                - 41 = MSECM
                - 42 = MSEFO
                - 43 = MSECD
                - 51 = MCXFO

            series (str): Series type, such as "FUTIDX", "FUTSTK", or other valid series codes.

            symbol (str): Trading symbol of the instrument, e.g., "NIFTY", "BANKNIFTY", "RELIANCE".

            expiryDate (str): Expiry date of the futures contract in the format "DDMMMYYYY", e.g., "26Jun2025".

        """
        try:
            params = {'exchangeSegment': exchangeSegment, 'series': series, 'symbol': symbol, 'expiryDate': expiryDate}
            response = await self._get('market.instruments.instrument.futuresymbol', params)
            return response
        except Exception as e:
            return response['description']

    async def get_option_symbol(
        self,
        exchangeSegment: int,
        series: str,
        symbol: str,
        expiryDate: str,
        optionType: str,
        strikePrice: float
        ):
        """
        Retrieves the option symbol for a given set of parameters from the specified exchange segment.

        Parameters:
            exchangeSegment (int): Numeric identifier for the exchange segment. Accepted values include:
                - 1  = NSECM
                - 2  = NSEFO
                - 3  = NSECD
                - 4  = NSECO
                - 5  = SLBM
                - 7  = NIFSC
                - 11 = BSECM
                - 12 = BSEFO
                - 13 = BSECD
                - 14 = BSECO
                - 21 = NCDEX
                - 41 = MSECM
                - 42 = MSEFO
                - 43 = MSECD
                - 51 = MCXFO

            series (str): Series type, such as "EQ", "FUT", "OPTIDX", "OPTSTK", "OPTFO".

            symbol (str): Trading symbol of the instrument, e.g., "RELIANCE", "TATAMOTORS", "NIFTY", "BANKNIFTY".

            expiryDate (str): Expiry date of the option in the format "DDMMMYYYY", e.g., "26Jun2025".

            optionType (str): Type of option - "CE" for Call Option or "PE" for Put Option.

            strikePrice (float): Strike price of the option, e.g., 24500.0.

        """
        try:
            params = {'exchangeSegment': exchangeSegment, 'series': series, 'symbol': symbol, 'expiryDate': expiryDate,
                      'optionType': optionType, 'strikePrice': strikePrice}
            response = await self._get('market.instruments.instrument.optionsymbol', params)
            return response
        except Exception as e:
            return response['description']

    async def get_option_type(
        self,
        exchangeSegment: int,
        series: str,
        symbol: str,
        expiryDate: str):
        """
        Retrieves the available option types (Call/Put) for a given instrument and expiry date 
        on the specified exchange segment.

        Parameters:
            exchangeSegment (int): Numeric identifier for the exchange segment. Accepted values include:
                - 1  = NSECM
                - 2  = NSEFO
                - 3  = NSECD
                - 4  = NSECO
                - 5  = SLBM
                - 7  = NIFSC
                - 11 = BSECM
                - 12 = BSEFO
                - 13 = BSECD
                - 14 = BSECO
                - 21 = NCDEX
                - 41 = MSECM
                - 42 = MSEFO
                - 43 = MSECD
                - 51 = MCXFO

            series (str): Series type, such as "OPTIDX", "OPTSTK", or similar.

            symbol (str): Trading symbol of the instrument, e.g., "NIFTY", "BANKNIFTY", "RELIANCE".

            expiryDate (str): Expiry date in the format "DDMMMYYYY", e.g., "26Jun2025".

            """
        try:
            params = {'exchangeSegment': exchangeSegment, 'series': series, 'symbol': symbol, 'expiryDate': expiryDate}
            response = await self._get('market.instruments.instrument.optiontype', params)
            return response
        except Exception as e:
            return response['description']

    async def get_index_list(self, exchangeSegment):
        """ Get the index list of the exchange segment."""
        try:
            params = {'exchangeSegment': exchangeSegment}
            response = await self._get('market.instruments.indexlist', params)
            return response
        except Exception as e:
            return response['description']

    async def search_by_instrumentid(self, Instruments):
        """ Search by instrument id."""
        try:
            params = {'source': self.source, 'instruments': Instruments}
            response = await self._post('market.search.instrumentsbyid', json.dumps(params))
            return response
        except Exception as e:
            return response['description']

    async def search_by_scriptname(self, searchString):
        """ Search by script name."""
        try:
            params = {'searchString': searchString}
            response = await self._get('market.search.instrumentsbystring', params)
            return response
        except Exception as e:
            return response['description']

    async def marketdata_logout(self):
        """This call invalidates the session token and destroys the API session. After this, the user should go"""
        try:
            params = {}
            response = await self._delete('market.logout', params)
            self.token = None # Added this to reset token, so that we can login again.
            return response
        except Exception as e:
            return response['description']

    ########################################################################################################
    # Common Methods
    ########################################################################################################

    async def _get(self, route, params=None):
        """Alias for sending a GET request."""
        return await self._request(route, "GET", params)

    async def _post(self, route, params=None):
        """Alias for sending a POST request."""
        return await self._request(route, "POST", params)

    async def _put(self, route, params=None):
        """Alias for sending a PUT request."""
        return await self._request(route, "PUT", params)

    async def _delete(self, route, params=None):
        """Alias for sending a DELETE request."""
        return await self._request(route, "DELETE", params)

    async def _request(self, route, method, parameters=None):
        """Make an HTTP request."""
        params = parameters if parameters else {}

        # Form a restful URL
        uri = self._routes[route].format(params)
        url = urljoin(self.root, uri)
        headers = {}

        if self.token:
            # set authorization header
            headers.update({'Content-Type': 'application/json', 'Authorization': self.token})

        try:
            # r = await self.reqsession.request(method,
            #                             url,
            #                             data=params if method in ["POST", "PUT"] else None,
            #                             params=params if method in ["GET", "DELETE"] else None,
            #                             headers=headers,
            #                             verify=not self.disable_ssl)

            r = await self.reqsession.request(method = method, url = url,
                                        data=params if method in ["POST", "PUT"] else None,
                                        params=params if method in ["GET", "DELETE"] else None,
                                        headers=headers)
        except Exception as e:
            raise e

        if self.debug:
            log.debug("Response: {code} {content}".format(code=r.status_code, content=r.content))

        # Validate the content type.
        if "json" in r.headers["content-type"]:
            try:
                data = json.loads(r.content.decode("utf8"))
            except ValueError:
                raise ex.XTSDataException("Couldn't parse the JSON response received from the server: {content}".format(
                    content=r.content))

            # api error
            if data.get("type"):

                if r.status_code == 400 and data["type"] == "error" and data["description"] == "Invalid Token":
                    raise ex.XTSTokenException(data["description"])

                if r.status_code == 400 and data["type"] == "error" and data["description"] == "Bad Request":
                    message = "Description: " + data["description"] + " errors: " + str(data['result']["errors"])
                    raise ex.XTSInputException(str(message))

            return data
        else:
            raise ex.XTSDataException("Unknown Content-Type ({content_type}) with response: ({content})".format(
                content_type=r.headers["content-type"],
                content=r.content))
